/******************************************************************************
 *
 * Project:  GDAL
 * Purpose:  Return the path of the "gdal" binary
 * Author:   Even Rouault <even dot rouault at spatialys.com>
 *
 ******************************************************************************
 * Copyright (c) 2025, Even Rouault <even dot rouault at spatialys.com>
 *
 * SPDX-License-Identifier: MIT
 ****************************************************************************/

#ifndef GDALGETGDALPATH_INCLUDED_H
#define GDALGETGDALPATH_INCLUDED_H

#include <string>

std::string GDALGetGDALPath();

#endif  // GDALGETGDALPATH_INCLUDED_H
